import logging

from enoslib import set_config
from rich.logging import RichHandler

from e2clab.constants import (
    LOG_E2CLAB_PREFIX,
    LOG_ERR_FILENAME,
    LOG_INFO_FILENAME,
    LOG_WRITING_MODE,
)


class E2clabLogger(logging.LoggerAdapter):
    """
    Custom logger adapter to add prefixes to E2clab logs for better readability.
    """

    def process(self, msg, kwargs):
        extra_tags = self.extra["tags"]
        if extra_tags is not None:
            prefix = ",".join([LOG_E2CLAB_PREFIX, *self.extra["tags"]])
        else:
            prefix = LOG_E2CLAB_PREFIX
        return f"[{prefix}] {msg}", kwargs


def getLogger(name, tags=None) -> E2clabLogger:
    """
    Returns a E2clab logger
    """
    if tags is None:
        tags = None
    logger = E2clabLogger(logging.getLogger(name), dict(tags=tags))
    return logger


def init_logging(
    level=logging.INFO,
    enable_enoslib: bool = True,
    mute_ansible: bool = False,
    file_handler: bool = True,
    **kwargs,
):
    """
    Initiates the global loggers for :
        - E2clab
        - EnOSlib
    Sets terminal and file handler for root logger.
    """
    default_kwargs = dict(
        show_time=False,
        rich_tracebacks=True,
    )

    default_kwargs.update(**kwargs)

    # Setup global parent logger for e2clab
    e2c_logger = logging.getLogger("e2clab")
    e2c_logger.setLevel(level=level)

    rich_handler = RichHandler(**default_kwargs)
    rich_formatter = logging.Formatter("%(message)s")
    rich_handler.setFormatter(rich_formatter)

    if file_handler:
        FILE_FORMATTER = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        file_formatter = logging.Formatter(FILE_FORMATTER)
        # Overwrites logging file by default
        file_handler_info = logging.FileHandler(LOG_INFO_FILENAME, LOG_WRITING_MODE)
        file_handler_info.setLevel(level=level)
        file_handler_info.setFormatter(file_formatter)

        file_handler_error = logging.FileHandler(LOG_ERR_FILENAME, LOG_WRITING_MODE)
        file_handler_error.setLevel(level=logging.ERROR)
        file_handler_error.setFormatter(file_formatter)

    # If we want to see logs from EnOSlib, we need to setup the logger
    # We don't use the enoslib.init_logging method as it directly setups
    # the root logger and gives us less fine-tuning.
    if enable_enoslib:
        en_logger = logging.getLogger("enoslib")
        en_logger.setLevel(level=level)

    if mute_ansible:
        set_config(ansible_stdout="noop")

    # Setup the root logger for terminal and file output
    root_logger = logging.getLogger()
    root_logger.addHandler(rich_handler)
    if file_handler:
        root_logger.addHandler(file_handler_info)
        root_logger.addHandler(file_handler_error)
